# © 2020 initOS GmbH
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl).

import logging

_logger = logging.getLogger(__name__)


class View:
    def __init__(self, name=None, inherit=None, calls=None, model=None):
        self.name = name
        self.inherit = inherit
        self.calls = set(calls or [])
        self.model = model

    def __repr__(self):
        return "<View: %s>" % self.name

    def copy(self):
        return View(self.name, self.inherit, self.calls.copy())

    def update(self, other):
        if self.name == other.name:
            self.calls.update(other.calls)

    @staticmethod
    def enforce_fullname(name, module_name):
        if not isinstance(name, str):
            return None
        if "." in name:
            return name
        return "%s.%s" % (module_name, name)

    def to_json(self):
        return {
            "name": self.name,
            "inherit": self.inherit,
            "model": self.model,
            "calls": list(self.calls),
        }

    @classmethod
    def from_json(cls, data):
        return cls(
            name=data.get("name", None),
            inherit=data.get("inherit", None),
            model=data.get("models", None),
            calls=data.get("calls", None),
        )

    @classmethod
    def from_xml(cls, module_name, obj):
        name = obj.attrib.get("id", None)

        if obj.tag == "template":
            inherit = obj.attrib.get("inherit_id", None)
            calls = obj.xpath(".//@t-call")
            model = None
        elif obj.tag == "record" and obj.attrib.get("model") == "ir.ui.view":
            tmp = obj.xpath('./field[@name="inherit_id"]/@ref')
            inherit = tmp[0] if tmp else None
            calls = obj.xpath('./field[@name="arch"]//@t-call')
            model = obj.xpath('./field[@name="model"]//text()')
            model = model[0] if model else None
        else:
            return None

        calls = {View.enforce_fullname(c, module_name) for c in calls}
        name = View.enforce_fullname(name, module_name)

        if name:
            return cls(name, inherit, calls, model)
        if isinstance(inherit, str):
            name = "%s.%s" % (module_name, inherit.rsplit(".")[-1])
            return cls(name, inherit, calls, model)
        return None



# © 2020 initOS GmbH
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl).

import ast


class Model:
    def __init__(self, name=None, inherit=None, inherits=None, fields=None):
        self.name = name
        self.inherit = set(inherit or [])
        self.inherits = inherits or {}
        self.fields = fields or {}

    def __repr__(self):
        return "<Model: %s>" % self.name

    def copy(self):
        return Model(
            self.name, self.inherit.copy(), self.inherits.copy(), self.fields.copy(),
        )

    def update(self, other):
        if self.name == other.name:
            self.inherit.update(other.inherit)
            self.inherits.update(other.inherits)
            self.fields.update(other.fields)

    def _parse_assign(self, obj):
        assignments = [k.id for k in obj.targets if isinstance(k, ast.Name)]
        if len(assignments) != 1:
            return

        assign, value = assignments[0], obj.value
        try:
            if assign == "_name":
                self.name = ast.literal_eval(value)
            elif assign == "_inherit":
                value = ast.literal_eval(value)
                if isinstance(value, list):
                    self.inherit.update(value)
                else:
                    self.inherit.add(value)
            elif assign == "_inherits":
                inhs = ast.literal_eval(value)
                if isinstance(inhs, dict):
                    self.inherits.update(inhs)
                    self.fields.update({k: "fields.Many2one" for k in inhs})
            elif isinstance(value, ast.Call):
                f = value.func
                if not isinstance(f, ast.Attribute):
                    return

                if f.value.id == "fields":
                    self.fields[assign] = "fields.%s" % f.attr

        except Exception:
            pass

    def to_json(self):
        return {
            "name": self.name,
            "inherit": list(self.inherit),
            "inherits": self.inherits,
            "fields": self.fields,
        }

    @classmethod
    def from_json(cls, data):
        return cls(
            name=data.get("name", None),
            inherit=data.get("inherit", None),
            inherits=data.get("inherits", None),
            fields=data.get("fields", None),
        )

    @classmethod
    def from_ast(cls, obj):
        model = cls()
        for child in obj.body:
            if isinstance(child, ast.Assign):
                model._parse_assign(child)

        if model.name:
            return model
        if len(model.inherit) == 1:
            model.name = list(model.inherit)[0]
            return model
        return None



# © 2020 initOS GmbH
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl).

from .model import Model
from .module import Module
from .odoo import Odoo
from .utils import folder_blacklist, stopwords
from .view import View

__all__ = ["Model", "Module", "Odoo", "folder_blacklist", "stopwords", "View"]



# © 2020 initOS GmbH
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl).

import json
import logging
import os
from configparser import ConfigParser
from fnmatch import fnmatch
from functools import reduce

try:
    from graphviz import Digraph, Graph
except ImportError:
    Digraph = Graph = None

from .module import Module

_logger = logging.getLogger(__name__)


class Odoo:
    def __init__(self, config="analyse.cfg", modules=None, filters=None):
        self.load_config(config)
        self.full = modules or {}
        self.modules = modules or {}
        self.filters = filters or []

    @classmethod
    def from_config(cls, config_path):
        if not os.path.isfile(config_path):
            return None

        odoo = cls()
        odoo.load(config_path)
        return odoo

    @classmethod
    def from_path(cls, path):
        if not os.path.isdir(path):
            return None

        odoo = cls()
        odoo.load_path(path)
        return odoo

    def __len__(self):
        return len(self.modules)

    def __iter__(self):
        return iter(self.modules)

    def __contains__(self, name):
        return name in self.modules

    def __getitem__(self, name):
        return self.modules[name]

    def load_config(self, file):
        self.config = {}
        if not os.path.isfile(file):
            return

        cp = ConfigParser()
        cp.read(file)

        for section_name, section in cp.items():
            self.config[section_name] = dict(section)
            for option_name, value in section.items():
                key = "%s.%s" % (section_name, option_name)
                self.config[key] = value

    def set_opt(self, option, value):
        self.config[option] = value

    def opt(self, name, default=None):
        return self.config.get(name, default) or None

    def items(self):
        return self.modules.items()

    def models(self):
        res = {}
        for module in self.modules.values():
            for model in module.models.values():
                if model.name in res:
                    res[model.name].update(model)
                else:
                    res[model.name] = model.copy()
        return res

    def views(self):
        res = {}
        for module in self.modules.values():
            for view in module.views.values():
                if view.name in res:
                    res[view.name].update(view)
                else:
                    res[view.name] = view.copy()
        return res

    def test_filter(self):
        _logger.debug("Applying filter: test")
        self.modules = {
            name: module
            for name, module in self.items()
            if not name.startswith("test_")
        }

    def path_filter(self, pattern):
        _logger.debug("Applying filter: path [%s]", pattern)
        self.modules = {
            name: module
            for name, module in self.items()
            if fnmatch(module.path, pattern)
        }

    def name_filter(self, pattern):
        _logger.debug("Applying filter: name [%s]", pattern)
        self.modules = {
            name: module for name, module in self.items() if fnmatch(name, pattern)
        }

    def load(self, config_path):
        _logger.debug("Reading odoo configuration file")
        cp = ConfigParser()
        cp.read(config_path)

        paths = cp.get("options", "addons_path", fallback="").split(",")
        self.load_path(paths, depth=1)

    def _full_dependency(self, name):
        if name not in self:
            return set()

        res = set()
        mods = list(self[name].depends)
        while mods:
            mod = mods.pop()
            if mod not in res and mod in self.full:
                res.add(mod)
                mods.extend(self.full[mod].depends)
        return res

    def interactive(self):
        local_vars = {"analyse": self}
        try:
            from IPython import start_ipython

            start_ipython(argv=[], user_ns=local_vars)
        except ImportError:
            from code import interact

            banner = ["%s: %s" % pair for pair in sorted(local_vars.items())]
            interact("\n".join(banner), local=local_vars)

    def analyse(self, file_path):
        _logger.debug("Start analysing...")
        models = {
            mname: name
            for name, module in self.full.items()
            for mname, model in module.models.items()
            if not model.inherit and not model.inherits
        }

        res = {}
        for name, module in self.items():
            fields = 0
            used = module.imports.union(module.refers)
            for model in module.models.values():
                fields += len(model.fields)

                used.union({models[x] for x in model.inherit if x in models})
                used.union({models[x] for x in model.inherits if x in models})

            full = self._full_dependency(name)
            missing = used.difference(full)
            missing.discard("base")

            res[name] = {
                "data_count": module.data,
                "depends": sorted(module.depends),
                "fields": fields,
                "imports": sorted(module.imports),
                "model_count": len(module.models),
                "refers": sorted(used),
            }
            if missing:
                _logger.error("Missing dependency: %s -> %s", name, missing)
                res[name]["missing_dependency"] = sorted(missing)

        if file_path == "-":
            print(json.dumps(res, indent=2))
        else:
            with open(file_path, "w+") as fp:
                json.dump(res, fp, indent=2)

    def load_path(self, paths, depth=None):
        if isinstance(paths, str):
            paths = [paths]

        result = Module.find_modules(paths, depth=depth)

        self.full = result.copy()
        self.modules = result.copy()

    def load_json(self, filename):
        with open(filename) as fp:
            data = json.load(fp)

            self.modules = {k: Module.from_json(v) for k, v in data.items()}
            self.full = self.modules.copy()

    def save_json(self, filename):
        with open(filename, "w+") as fp:
            json.dump({k: m.to_json() for k, m in self.full.items()}, fp)

    def _find_edges_in_loop(self, graph):
        # Eliminate not referenced and not referring modules
        while True:
            before = len(graph)

            tmp = {k: v.intersection(graph) for k, v in graph.items()}
            deps = reduce(lambda a, b: a.union(b), tmp.values(), set())
            graph = {k: v for k, v in tmp.items() if v and k in deps}

            after = len(graph)
            if before == after:
                break

        return {(a, b) for a, bs in graph.items() for b in bs}

    def _show_graph(
        self, graph, node_check=None, color_node=None, color_edge=None, filename=None,
    ):
        if not graph:
            return

        # Don't show nodes without dependencies
        if self.opt("odoo.skip_lonely_node", True):
            depends = reduce(lambda a, b: a.union(b), graph.values(), set())
            depends.intersection_update(graph)

            tmp, graph = graph.copy(), {}
            for name, deps in tmp.items():
                deps.intersection_update(depends)
                if deps or name in depends:
                    graph[name] = deps

        # Set the visible nodes
        if callable(node_check):
            visible = set(filter(node_check, graph))
        else:
            visible = set(graph)

        # Show all dependency ignoring the filters
        if self.opt("odoo.show_full_dependency"):
            nodes = list(visible)
            visited = set()
            while nodes:
                current = nodes.pop()
                if current in visited:
                    continue
                visited.add(current)

                depends = graph[current]
                visible.update(depends)
                nodes.extend(depends)

        if not visible:
            return

        output = Digraph(engine=self.opt("odoo.engine", "dot"))
        for name in visible:
            depends = graph[name]
            if callable(color_node):
                output.node(name, color=color_node(name))
            else:
                output.node(name)

            for dep in depends:
                if dep not in visible:
                    continue

                if callable(color_edge):
                    output.edge(name, dep, color=color_edge(name, dep))
                else:
                    output.edge(name, dep)

        self._show_output(output, filename=filename)

    def _show_output(self, graph, filename):
        graph.view(filename=filename)

    def show_structure_graph(
        self, modules="*", models="*", views="*", fields=True, filename=None
    ):
        module_color = self.opt("structure.module_color")
        module_shape = self.opt("structure.module_shape", "doubleoctagon")
        model_color = self.opt("structure.model_color")
        model_shape = self.opt("structure.model_shape", "box")
        view_color = self.opt("structure.view_color")
        view_shape = self.opt("structure.view_shape", "oval")
        field_color = self.opt("structure.field_color")
        field_shape = self.opt("structure.field_shape", "octagon")

        output = Graph()

        def render_field(model_id, model):
            for field_name in model.fields:
                field_id = "%s/%s" % (model_id, field_name)
                output.node(
                    field_id, label=field_name, color=field_color, shape=field_shape
                )
                output.edge(model_id, field_id)

        def render_model(module_id, model_name, model):
            model_id = "%s/%s" % (module_id, model_name)
            output.node(
                model_id, label=model_name, color=model_color, shape=model_shape
            )
            output.edge(module_id, model_id)

            if fields:
                render_field(model_id, model)

        def render_view(module_id, view_name, view):
            view_id = "%s/%s" % (module_id, view_name)
            output.node(view_id, label=view_name, color=view_color, shape=view_shape)
            output.edge(module_id, view_id)

        for module_name, module in self.items():
            if not fnmatch(module_name, modules):
                continue

            module_id = module_name
            output.node(
                module_id, label=module_name, color=module_color, shape=module_shape,
            )

            for model_name, model in module.models.items():
                if fnmatch(model_name, models):
                    render_model(module_id, model_name, model)

            for view_name, view in module.views.items():
                if fnmatch(view_name, views):
                    render_view(module_id, view_name, view)

        self._show_output(output, filename=filename or "structure.gv")

    def show_module_graph(
        self,
        modules="*",
        version=False,
        depends=True,
        imports=False,
        refers=False,
        filename=None,
    ):
        # Build the dependency graph
        graph = {}
        for name, module in self.items():
            graph[name] = set()
            if depends:
                graph[name].update(module.depends)
            if imports:
                graph[name].update(module.imports)
            if refers:
                graph[name].update(module.refers)

        # Detect loops
        loop_edges = self._find_edges_in_loop(graph)
        # Evaluate the migration state if possible
        if version:
            migrated = {name for name, module in self.items()}
            fully_migrated = {
                name
                for name, mod in self.items()
                if all(dep in migrated for dep in mod.depends if dep in self)
            }
        else:
            migrated = fully_migrated = set()

        # Options
        done_color = self.opt("module.done_color", "green")
        todo_color = self.opt("module.todo_color", "blue")
        warning_color = self.opt("module.warning_color", "orange")
        default_color = self.opt("module.default_color")
        loop_color = self.opt("module.loop_color", "red")

        def check_node(node):
            return fnmatch(node, modules)

        # Coloring functions
        def color_node(node):
            if node in migrated and node in fully_migrated:
                return done_color
            if node in fully_migrated:
                return todo_color
            if node in migrated:
                return warning_color
            return default_color

        def color_edge(node_a, node_b):
            if (node_a, node_b) in loop_edges:
                return loop_color
            return None

        # Show the resulting graph
        self._show_graph(
            graph, check_node, color_node, color_edge, filename=filename or "module.gv",
        )

    def show_model_graph(self, models="*", inherit=True, inherits=True, filename=None):
        graph = {}
        for name, model in self.models().items():
            graph[name] = set()

            if inherit:
                graph[name].update(model.inherit)
            if inherits:
                graph[name].update(model.inherits)

        # Detect loops
        loop_edges = self._find_edges_in_loop(graph)

        # Options
        base_color = self.opt("model.base_color", "blue")
        default_color = self.opt("model.default_color")
        loop_color = self.opt("model.loop_color", "red")

        def check_node(node):
            return fnmatch(node, models)

        # Coloring functions
        def color_node(node):
            if not graph.get(node, True):
                return base_color
            return default_color

        def color_edge(node_a, node_b):
            if (node_a, node_b) in loop_edges:
                return loop_color
            return None

        self._show_graph(
            graph, check_node, color_node, color_edge, filename=filename or "model.gv"
        )

    def show_view_graph(self, views="*", inherit=True, calls=True, filename=None):
        """Show the graph of the views"""
        graph = {}
        for name, view in self.views().items():
            graph[name] = set()

            if inherit:
                graph[name].add(view.inherit)
            if calls:
                graph[name].update(view.calls)

        # Detect loops
        loop_edges = self._find_edges_in_loop(graph)

        # Options
        base_color = self.opt("view.base_color", "blue")
        default_color = self.opt("view.default_color")
        loop_color = self.opt("view.loop_color", "red")

        def check_node(node):
            return fnmatch(node, views)

        # Coloring functions
        def color_node(node):
            if not graph.get(node, True):
                return base_color
            return default_color

        def color_edge(node_a, node_b):
            if (node_a, node_b) in loop_edges:
                return loop_color
            return None

        self._show_graph(
            graph, check_node, color_node, color_edge, filename=filename or "view.gv"
        )



# © 2019 initOS GmbH
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl).

import hashlib
import logging
import os
import re
import shutil
import subprocess

_logger = logging.getLogger(__name__)


def call(cmd, cwd=None):
    proc = subprocess.Popen(
        cmd,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
        cwd=cwd,
        universal_newlines=True,
        env={**os.environ, "GIT_TERMINAL_PROMPT": "0"},
    )
    return [pipe.strip() for pipe in proc.communicate()]


def stopwords(words=None):
    """ Returns or stores the stopword set """
    if words:
        stopwords.words = words
        return words
    if hasattr(stopwords, "words"):
        return stopwords.words
    return set()


def folder_blacklist(blacklist=None):
    """ Returns or stores the folder blacklist """
    if blacklist:
        folder_blacklist.blacklist = blacklist
        return blacklist
    if hasattr(folder_blacklist, "blacklist"):
        return folder_blacklist.blacklist
    return set()


def hexhash(s):
    """ Generates a hash and returns it as hex """
    if not isinstance(s, bytes):
        s = s.encode()
    return hashlib.md5(s).hexdigest()


def hexhash_files(files):
    """ Generates a hash for a list of files """
    hashes = []
    for f in sorted(set(files)):
        hashsum = hexhash(open(f, "rb").read())
        hashes.append(f"{hashsum} {f}")
    return hexhash("\n".join(hashes))


def fix_indentation(filepath):
    """ Fixes the indentation of a file """
    result = False
    with open(filepath, "r+") as fp:
        buf = fp.read()

    with open(filepath, "w+") as fp:
        for line in buf.splitlines():
            left = ""
            for c in line:
                if c == " ":
                    left += c
                elif c == "\t":
                    left += " " * (4 - len(left) % 4)
                    result = True
                else:
                    break
            fp.write(f"{left}{line.strip()}\n")

    return result


def try_automatic_port(filepath):
    """ Tries to port a python 2 script to python 3 using 2to3 """
    cmd = shutil.which("2to3")
    if cmd is None:
        _logger.warning("Automatic porting needs 2to3 installed")
        return False

    proc = subprocess.Popen(
        [cmd, "-n", "-w", filepath], stdout=subprocess.PIPE, stderr=subprocess.PIPE,
    )
    proc.communicate()
    return True


def analyse_language(path):
    """ Analyse the languages of a directory """
    cmd = shutil.which("cloc")
    if cmd is None:
        _logger.warning("Language analyse needs cloc")
        return {}

    output, error = call([cmd, path])
    if error:
        return {}

    result = {}
    reg = re.compile(r"([^:]*?)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)")
    total_code = 0
    for line in output.splitlines():
        match = reg.match(line)
        if match:
            lang, _, _, _, code = match.groups()
            total_code += int(code)
            result[lang] = {"lines": int(code), "fraction_from_total": 0}

    for value in result.values():
        value["fraction_from_total"] = value["lines"] / total_code

    return result



# © 2020 initOS GmbH
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl).

import ast
import csv
import logging
import os
import re
import tempfile

from lxml import etree

from .model import Model
from .utils import (
    analyse_language,
    fix_indentation,
    folder_blacklist,
    hexhash_files,
    stopwords,
    try_automatic_port,
)
from .view import View

Manifests = {"__odoo__.py", "__openerp__.py", "__manifest__.py"}
_logger = logging.getLogger(__name__)


def is_readme(filename):
    return any(filename.startswith(k) for k in ("readme.", "README."))


class Module:
    def __init__(self, path):
        # Path to the module
        self.path = path
        # Technical name of the module
        self.name = path.rstrip("/").split("/")[-1]
        # Manifest of the module
        self.manifest = {}
        # Models defined in the module
        self.models = {}
        # Views defined in the module
        self.views = {}
        # Records defined in the module
        self.data = 0
        # Modules this module depends on
        self.depends = set()
        # Modules this module imports in python files
        self.imports = set()
        # Modules this module refers to in xml files
        self.refers = set()
        # Parsed files of the module
        self.files = set()
        # Status
        self.status = set()
        self.language = {}
        # readme file, status and description
        self.words = set()
        # hash of all the modules
        self.hashsum = ""

    @property
    def author(self):
        return self.manifest.get("author", "")

    @property
    def category(self):
        return self.manifest.get("category", "")

    @property
    def description(self):
        return self.manifest.get("description", "")

    @property
    def external_dependencies(self):
        return self.manifest.get("external_dependencies", {})

    @property
    def installable(self):
        return self.manifest.get("installable", "")

    @property
    def license(self):
        return self.manifest.get("license", "")

    @property
    def readme(self):
        for f in os.listdir(self.path):
            if is_readme(f):
                p = os.path.join(self.path, f)
                return open(p, "r").read()
        return ""

    @property
    def readme_type(self):
        for f in os.listdir(self.path):
            if is_readme(f):
                return os.path.splitext(f)[1] or None
        return None

    @property
    def summary(self):
        return self.manifest.get("summary", "")

    @property
    def version(self):
        return self.manifest.get("version", "")

    @property
    def website(self):
        return self.manifest.get("website", "")

    @property
    def info(self):
        deps = self.depends.union(self.imports).union(self.refers)
        return {
            "name": self.name,
            "model_count": len(self.models),
            "data_count": self.data,
            "depends": len(deps),
        }

    def add(self, **kwargs):
        if "depends" in kwargs:
            self.depends.add(kwargs["depends"])
        if "files" in kwargs:
            self.files.add(kwargs["files"])
        if "imports" in kwargs:
            self.imports.add(kwargs["imports"])
        if "refers" in kwargs:
            self.refers.add(kwargs["refers"])

    def update(self, **kwargs):
        self.models.update(kwargs.get("models", {}))
        self.depends.update(kwargs.get("depends", []))
        self.files.update(kwargs.get("files", []))
        self.imports.update(kwargs.get("imports", []))
        self.refers.update(kwargs.get("refers", []))

    def __repr__(self):
        return "<Module: %s>" % self.name

    def analyse_language(self):
        self.language = analyse_language(self.path)

    def analyse_hash(self, files_list):
        self.hashsum = hexhash_files(files_list)

    def _load_python(self, path, filename):
        def parse_python(path, filename):
            name = os.path.join(path, filename)
            with open(name) as fp:
                return ast.parse(fp.read())

        try:
            return parse_python(path, filename)
        except (SyntaxError, TabError):
            filepath = os.path.join(path, filename)
            with tempfile.NamedTemporaryFile("w+") as tmp:
                tmp.file.write(open(filepath, "r").read())
                if try_automatic_port(tmp.name):
                    _logger.warning("Ported %s", filepath)
                    self.status.add("ported")
                if fix_indentation(tmp.name):
                    _logger.warning("Fixed indentation %s", filepath)
                    self.status.add("indent-fix")
                return parse_python(*os.path.split(tmp.name))

    def _parse_class_def(self, obj):
        model = Model.from_ast(obj)
        if not model or not model.name:
            return
        if model.name in self.models:
            self.models[model.name].update(model)
        else:
            self.models[model.name] = model

    def _parse_python(self, path, filename):
        obj = self._load_python(path, filename)

        self.add(files=path + filename)

        imports = set()
        fmt = "{}.{}".format
        for c in obj.body:
            if isinstance(c, ast.ImportFrom):
                m = c.module
                imports.update(fmt(m or "", name.name) for name in c.names)
            elif isinstance(c, ast.Import):
                imports.update(name.name for name in c.names)

        for child in obj.body:
            if isinstance(child, ast.ClassDef):
                self._parse_class_def(child)

        patterns = ["odoo.addons.", "openerp.addons."]
        for imp in imports:
            if any(imp.startswith(p) for p in patterns):
                mod = imp.split(".")[2]
                if mod != self.name:
                    self.imports.add(mod)
                continue
            elif imp.split(".", 1)[0] in ["odoo", "openerp"]:
                continue

            p = path
            for f in imp.lstrip(".").split("."):
                if os.path.isfile("%s%s.py" % (p, f)):
                    self._parse_python(p, f + ".py")
                elif os.path.isfile("%s%s/__init__.py" % (p, f)):
                    self._parse_python("%s%s/" % (p, f), "__init__.py")
                elif os.path.isdir(p + f):
                    p += f + "/"
                else:
                    break

    def _parse_manifest(self, path):
        with open(path) as fp:
            obj = ast.literal_eval(fp.read())
            if isinstance(obj, dict):
                self.update(depends=obj.get("depends", []), files=obj.get("data", []))
                self.manifest.update(obj)
                self._parse_text_for_keywords(
                    [self.name, self.summary, self.description]
                )

    def _parse_csv(self, path):
        with open(path) as fp:
            obj = csv.reader(fp)
            self.data += max(0, obj.line_num - 1)

    def _parse_xml(self, path):
        obj = etree.parse(path)
        # Supported special tags defining data
        tags = [
            "act_window",
            "assert",
            "delete",
            "function",
            "menuitem",
            "record",
            "report",
            "template",
            "url",
        ]
        self.data += sum(len(obj.xpath("//%s" % tag)) for tag in tags)

        # xpaths to get all referred modules
        xpaths = [
            "//record/field[@name='inherit_id']/@ref",
            "//template/@inherit_id",
            "//record[@model='ir.ui.view']/field[@name='arch']//@t-call",
            "//template//@t-call",
        ]
        xpaths.extend("//%s/@id" % tag for tag in tags)

        xmlid = re.compile(r"\w+\.\w+")
        xpaths = " | ".join(xpaths)
        refs = {x.split(".")[0] for x in obj.xpath(xpaths) if xmlid.match(x)}
        self.refers.update({x for x in refs if x != self.name})

        # xpaths to extract views
        for node in obj.xpath("//record[@model='ir.ui.view'] | //template"):
            view = View.from_xml(self.name, node)
            if not view:
                continue
            elif view.name in self.views:
                self.views[view.name].update(view)
            else:
                self.views[view.name] = view

    def _parse_text_for_keywords(self, texts):
        if not isinstance(texts, list):
            texts = [texts]

        words = stopwords()
        for text in texts:
            tmp = {w.lower() for w in re.findall(r"\b[a-zA-Z]{2,}\b", text)}
            self.words |= tmp.difference(words)

    def _parse_readme(self, path):
        with open(path) as fp:
            self._parse_text_for_keywords(fp.read())

    def to_json(self):
        return {
            "path": self.path,
            "name": self.name,
            "manifest": self.manifest,
            "models": {n: m.to_json() for n, m in self.models.items()},
            "views": {n: v.to_json() for n, v in self.views.items()},
            "data": self.data,
            "depends": list(self.depends),
            "imports": list(self.imports),
            "refers": list(self.refers),
            "files": list(self.files),
            "status": list(self.status),
            "language": self.language,
            "words": list(self.words),
            "hashsum": self.hashsum,
        }

    @classmethod
    def from_json(cls, data):
        module = cls(data.get("path"))
        module.name = data["name"]
        module.manifest = data["manifest"]
        module.models = {n: Model.from_json(m) for n, m in data["models"].items()}
        module.views = {n: View.from_json(m) for n, m in data["views"].items()}
        module.data = data["data"]
        module.depends = set(data["depends"])
        module.imports = set(data["imports"])
        module.refers = set(data["refers"])
        module.files = set(data["files"])
        module.status = set(data["status"])
        module.language = data["language"]
        module.words = set(data["words"])
        module.hashsum = set(data["hashsum"])
        return module

    @classmethod
    def from_path(cls, path):
        files_list = []
        module = cls(path)
        found_init, found_manifest = False, False
        if not path.endswith("/"):
            path += "/"

        # Find the manifest scripts
        for f in Manifests:
            filepath = os.path.join(path, f)
            if os.path.isfile(filepath):
                found_manifest = True
                module._parse_manifest(filepath)

        if not found_manifest:
            return None

        for f in os.listdir(path):
            # Found the init script
            if f == "__init__.py":
                found_init = True
                module._parse_python(path, f)

            # Found the readme
            elif is_readme(f):
                module._parse_readme(path + f)

            filepath = os.path.join(path, f)
            if os.path.isfile(filepath):
                files_list.append(filepath)

        if not found_init:
            return None

        module.analyse_language()

        for file in module.files:
            file_path = os.path.join(path, file)
            files_list.append(file_path)
            if file.endswith(".xml"):
                module._parse_xml(file_path)
            elif file.endswith(".csv"):
                module._parse_csv(file_path)

        module.analyse_hash(files_list)

        _logger.info("Found module %s", module.name)
        if module.status:
            _logger.info("Status %s: %s", module.name, module.status)

        return module

    @classmethod
    def find_modules(cls, paths, depth=None):
        result = {}
        if isinstance(paths, str):
            paths = [paths]

        paths = [(p, 0) for p in paths]
        blacklist = folder_blacklist()
        # Breadth-first search
        while paths:
            path, d = paths.pop(0)
            if depth is not None and d > depth:
                continue

            try:
                module = cls.from_path(path)
            except Exception as e:
                _logger.exception(e)
                continue

            if module is not None:
                name = module.name
                if name not in result:
                    result[name] = module
            else:
                sub_paths = [
                    os.path.join(path, p)
                    for p in os.listdir(path)
                    if p not in blacklist
                ]
                paths.extend((p, d + 1) for p in sub_paths if os.path.isdir(p))

        return result



import os

from setuptools import find_packages, setup


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as f:
        return f.read()

setup(
    name="odoo-analyse",
    version="0.2",
    author="initOS GmbH",
    author_email="info@initos.com",
    description="Package to analyse odoo modules",
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    license="AGPL-3.0",
    keywords="odoo, modules, analyze, dependency, graph",
    url="https://github.com/initOS/odoo-analyse",
    packages=find_packages("src"),
    package_dir={"": "src"},
    package_name="odoo-analyse",
    include_package_data=True,
    entry_points={
        'console_scripts':
            ['odoo_analyse = odoo_analyse.main:main'],
    },
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*",
    install_requires=[
        "2to3",
        "cloc",
        "lxml",
    ],
    extras_require={
        "console": [
            "graphviz",
        ],
    },
    project_urls={
        'Documentation': 'https://github.com/initOS/odoo-analyse/blob/master/README.md',
        'Usage': 'https://odoo-code-search.com',
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Software Development",
    ],
)



#!/usr/bin/env python3
# © 2020 initOS GmbH
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl).

import argparse
import glob
import logging
import os
import sys

import graphviz

from .odoo import Odoo

_logger = logging.getLogger(__name__)


def parse_args():
    parser = argparse.ArgumentParser(
        description=(
            "Default color encoding for the module graph:\n"
            " * Green: Migrated module where all dependencies are migrated\n"
            " * Orange: Migrated module where not all dependencies are migrated\n"
            " * Blue: Not migrated module where all dependencies are migrated\n"
            " * Red: Edges belonging to a loop\n"
            "\n"
            "Default color encoding for the model/view graph:\n"
            " * Blue: Module without dependencies\n"
            " * Red: Edges belonging to a loop\n"
        ),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    group = parser.add_argument_group("Loading/Saving")
    group.add_argument(
        "-c",
        "--config",
        default=False,
        help="Specify an odoo configuration file to load modules",
    )
    group.add_argument(
        "-p",
        "--path",
        default=[],
        action="append",
        help="Specify a path to search for odoo modules",
    )
    group.add_argument(
        "-l", "--load", default=False, help="Load from a json file",
    )
    group.add_argument(
        "-s", "--save", default=False, help="Save to a json file",
    )

    group = parser.add_argument_group("Filters")
    group.add_argument(
        "--path-filter",
        default="*",
        help="Filter out modules which paths aren't matching the glob",
    )
    group.add_argument(
        "--models",
        default="*",
        help="Filter out models which names aren't matching the glob",
    )
    group.add_argument(
        "--modules",
        default="*",
        help="Filter out modules which names aren't matching the glob",
    )
    group.add_argument(
        "--views",
        default="*",
        help="Filter out views which names aren't matching the glob",
    )
    group.add_argument(
        "--test-filter",
        action="store_true",
        default=False,
        help="Include testing modules starting with test_",
    )

    group = parser.add_argument_group(
        "Module graphs",
        "Generate a module dependency graph using the following options to "
        "Specify the visible dependencies",
    )
    group.add_argument(
        "--show-dependency",
        action="store_true",
        default=False,
        help="Show the module dependency of the manifest",
    )
    group.add_argument(
        "--show-import",
        action="store_true",
        default=False,
        help="Show python imports between modules",
    )
    group.add_argument(
        "--show-reference",
        action="store_true",
        default=False,
        help="Show xml references between modules",
    )
    group.add_argument(
        "--migration",
        default="*",
        help="Color the migration status in the module graph. "
        "Must be a glob which matches all migrated versions",
    )

    group = parser.add_argument_group("Model graph")
    group.add_argument(
        "--model-graph",
        action="store_true",
        default=False,
        help="Show the dependency graph of the models",
    )
    group.add_argument(
        "--no-model-inherit",
        action="store_true",
        default=False,
        help="Don't use inherit in the dependency graph of the models",
    )
    group.add_argument(
        "--no-model-inherits",
        action="store_true",
        default=False,
        help="Don't use inherits in the dependency graph of the models",
    )

    group = parser.add_argument_group("View graph")
    group.add_argument(
        "--view-graph",
        action="store_true",
        default=False,
        help="Show the dependency graph of the views",
    )
    group.add_argument(
        "--no-view-inherit",
        action="store_true",
        default=False,
        help="Don't use inherit in the dependency graph of the views",
    )
    group.add_argument(
        "--no-view-call",
        action="store_true",
        default=False,
        help="Don't use t-calls in the dependency graph of the views",
    )

    group = parser.add_argument_group("Stucture graph")
    group.add_argument(
        "--structure-graph",
        action="store_true",
        default=False,
        help="Show the structure of the modules",
    )

    group = parser.add_argument_group("Misc")
    group.add_argument(
        "--analyse",
        default="",
        help="Analyse the modules and store it in the given file",
    )
    group.add_argument(
        "-i",
        "--interactive",
        default=False,
        action="store_true",
        help="Enter the interactive mode",
    )

    group = parser.add_argument_group("Options")
    group.add_argument(
        "--full-graph",
        action="store_true",
        default=False,
        help="Show the full graph and only use the filters for the starting nodes",
    )
    group.add_argument(
        "--renderer",
        default="dot",
        help="Specify the rendering engine. %s" % graphviz.ENGINES,
    )

    return parser.parse_args()


def main():
    args = parse_args()

    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(logging.Formatter("%(message)s"))

    logger = logging.getLogger()
    logger.addHandler(handler)

    # Load modules
    if args.config:
        odoo = Odoo.from_config(args.config)
    else:
        odoo = Odoo()

    if args.interactive:
        odoo.interactive()
        sys.exit()

    if args.load:
        odoo.load_json(args.load)

    for p in args.path:
        odoo.load_path(glob.glob(os.path.expanduser(p)))

    # Save the modules
    if args.save:
        odoo.save_json(args.save)

    # Set global options
    if args.full_graph:
        odoo.show_full_dependency = True

    if args.renderer in graphviz.ENGINES:
        odoo.set_opt("odoo.engine", args.renderer)

    # Apply the filters
    if not args.test_filter:
        odoo.test_filter()
    odoo.path_filter(args.path_filter)

    if args.analyse:
        odoo.analyse(args.analyse)

    # Render the module graph if needed
    if args.dependency_graph or args.import_graph or args.reference_graph:
        odoo.show_module_graph(
            args.modules,
            args.migration,
            args.dependency_graph,
            args.import_graph,
            args.reference_graph,
        )

    # Render the structure graph
    if args.structure_graph:
        odoo.show_structure_graph(args.modules, args.models, args.views)

    # Render the model graph
    if args.model_graph:
        odoo.show_model_graph(
            args.models,
            inherit=not args.no_model_inherit,
            inherits=not args.no_model_inherits,
        )

    # Render the view graph
    if args.view_graph:
        odoo.show_view_graph(
            args.views, inherit=not args.no_view_inherit, calls=not args.no_view_call,
        )


if __name__ == "__main__":
    main()
